package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMotionWordSoftAssignment;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartTheme;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.DefaultXYDataset;
import xyz.matj1.bp.enums.MocapType;
import xyz.matj1.bp.experiments.KnnPrecision;
import xyz.matj1.bp.experiments.PairwiseCorrelation;
import xyz.matj1.bp.themes.SmoothXyGraphTheme;
import xyz.matj1.bp.themes.SuperimpositionXYGraphTheme;
import xyz.matj1.bp.types.PointSet;
import xyz.matj1.bp.types.Vector2;

import java.awt.Color;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.time.Duration;
import java.time.Instant;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static xyz.matj1.bp.Utils.loadData;
import static xyz.matj1.bp.enums.LengthRelation.QUOTIENT;

/**
 * A class for the testing of how sequences of
 * motion words of similar motions of different lengths can be compared
 *
 * @author Matěj Bagar
 */
public class LengthTesting {
	public static void main(String[] args) throws IOException{
		ObjectMotionWordSoftAssignment.maxPartsToMatch = 6;

//		pairwiseTests("output");
//		pairwiseTestsSmoothed("output/smoothed quotient pairwise graphs");
//		pairwiseTestsSmoothedWithQuantity("output/smoothed quotient pairwise graphs with quantity");
//		pairwiseTestsSuperimposed("output/superimposed quotient pairwise graphs");
//		knnTests("output");
//		knnTestsAveraged("output/kNN averages/kNN averages.txt");
//		knnTestsAveragedToGraphs("output/knn averages graphs");
		knnTestsAveragedOrderedToGraphs("output/knn averages ordered graphs");
	}

	/**
	 * Pairwise test for each category to graphs in a dedicated directory for each mocap type
	 *
	 * @param outputPath where the dedicated directories shall be
	 */
	private static void pairwiseTests(String outputPath) throws IOException {
		ShowGraph showQuotientGraph = (dataPath, dataClass, outputDirPath) -> {
			var timer = Instant.now();
			PairwiseCorrelation.show(
				loadData(dataPath, dataClass),
				QUOTIENT,
				new File(outputDirPath)
			);
			System.err.printf(
				"Time to compute pairwise correlelations for data class %s: %d s%n",
				dataClass.getSimpleName(),
				Duration.between(timer, Instant.now()).toSeconds()
			);
		};

		for (var type : MocapType.values()) {
			showQuotientGraph.show(
				dataFileNames(type),
				type.dataClass(),
				outputPath + "/" + type.adjective() + " quotient pairwise graphs"
			);
		}
	}

	/**
	 * For each mocap type, pairwise tests for each category
	 * with all data put into a single dataset and smoothed
	 *
	 * @param outputPath where the graphs shall be written
	 * @throws IOException if the reading or the writing fails
	 */
	private static void pairwiseTestsSmoothed(String outputPath) throws IOException {
		for (var type : MocapType.values()) {
			Utils.writeXYGraphAsPdf(
				PairwiseCorrelation.showCategory(
					PairwiseCorrelation.smooth(
						PairwiseCorrelation.compute(
							loadData(dataFileNames(type), type.dataClass()),
							QUOTIENT
						),
						0.01d,
						.5d
					),
					QUOTIENT,
					Optional.empty()
				),
				new File(outputPath, type.noun() + ".pdf"),
				new Vector2<>(512, 512),
				SmoothXyGraphTheme.instance()
			);
	}}

	/**
	 * Like smoothed pairwise graphs, but with an other data series showing how many series is in the x-area of the graph (relatively)
	 *
	 * @param outputPath where the graphs shall be written
	 * @throws IOException if the reading or the writing fails
	 */
	public static void pairwiseTestsSmoothedWithQuantity(String outputPath) throws IOException {
		// Both data series have the same color.
		// This needs relabeling the y axis.

		for (var type : MocapType.values()) {
			var pairwiseCorrelation = PairwiseCorrelation.compute(
				loadData(dataFileNames(type), type.dataClass()),
				QUOTIENT
			);

			var smoothCorrelation = PairwiseCorrelation.smooth(pairwiseCorrelation, 0.01, 0.5);
			var quantity = PairwiseCorrelation.quantity(pairwiseCorrelation, 0.01, 0.5);

			var quantityMaximum = quantity.stream().mapToDouble(Vector2::b).filter(Double::isFinite).max().orElseThrow();
			var sCMaximum = smoothCorrelation.stream().mapToDouble(Vector2::b).filter(Double::isFinite).max().orElseThrow();

			var quantityDataset = new DefaultXYDataset();

			// add quantity to the dataset
			double[] xQArray = quantity.stream().mapToDouble(Vector2::a).toArray();
			double[] yQArray = quantity.stream().mapToDouble(p -> p.b() * sCMaximum/quantityMaximum).toArray();
//			double[] yQArray = quantity.stream().mapToDouble(p -> p.b() / 1000).toArray();
			quantityDataset.addSeries(0, new double[][]{xQArray, yQArray});

			var chart = PairwiseCorrelation.showCategory(smoothCorrelation, QUOTIENT, Optional.empty());
			chart.getXYPlot().setDataset(1, quantityDataset);

			var theme = new ChartTheme() {
				@Override public void apply(JFreeChart chart) {
					SmoothXyGraphTheme.instance().apply(chart);
					var renderer = chart.getXYPlot().getRenderer();
					renderer.setSeriesPaint(1, Color.BLUE);
					renderer.setSeriesPaint(0, Color.GREEN);
			}};

			Utils.writeXYGraphAsPdf(chart, new File(outputPath, type.noun() + ".pdf"), new Vector2<>(512, 512), theme);
	}}

	/**
	 * Pairwise tests superimposed to one graph and written to a file for each mocap type
	 *
	 * @param outputPath where the graphs shall be written
	 * @throws IOException if the reading or the writing fails
	 */
	private static void pairwiseTestsSuperimposed(String outputPath) throws IOException {
		for (var type : MocapType.values()) {
			var data = loadData(dataFileNames(type), type.dataClass());
			var correlationData = PairwiseCorrelation.compute(data, QUOTIENT);

			var charts = data.getCategories().parallelStream().map(
				category -> PairwiseCorrelation.showCategory(correlationData.get(category), QUOTIENT, Optional.of(category))
			).toList();

			var superimposed = Utils.superimpose(charts);

			Utils.writeXYGraphAsPdf(
				superimposed,
				new File(outputPath, type.noun() + ".pdf"),
				new Vector2<>(512, 512),
				SuperimpositionXYGraphTheme.instance()
			);
	}}

	/**
	 * kNN tests with a graph for each category in a directory dedicated for each type
	 *
	 * @param outputPath where the dedicated directories shall be
	 * @throws IOException if the reading or the writing fails
	 */
	private static void knnTests(String outputPath) throws IOException {
		ShowGraph showKnnLengthGraph = (dataPath, dataClass, outputDirPath) -> {
			var timer = Instant.now();
			KnnPrecision.show(
				loadData(dataPath, dataClass),
				new File(outputDirPath)
			);
			System.err.printf(
				"Time to compute KNN for data class %s: %d s%n",
				dataClass.getSimpleName(),
				Duration.between(timer, Instant.now()).toSeconds()
			);
		};

		for (var type : MocapType.values()) {
			showKnnLengthGraph.show(
				dataFileNames(type),
				type.dataClass(),
				outputPath + "/" + type.adjective() + " kNN graphs"
			);
	}}

	/**
	 * kNN tests averaged and to a file for each mocap type
	 *
	 * @param outputPath where the result shall be written
	 * @throws IOException if the reading or the writing fails
	 */
	private static void knnTestsAveraged(String outputPath) throws IOException {
		var output = new PrintWriter(outputPath);

		for (var type : MocapType.values()) {
			output.printf("%n%s:%n", type.noun());
			Utils.write(
				KnnPrecision.averageFormat(
					KnnPrecision.average(KnnPrecision.compute(
						loadData(dataFileNames(type), type.dataClass())
				))),
				output
			);
	}}

	/**
	 * kNN tests averaged for each category, to a data series for each mocap type and to a graph
	 *
	 * @param outputPath where the result shall be written
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestsAveragedToGraphs(String outputPath) throws IOException {
		var typeAverages = new HashMap<MocapType, Map<Category, Double>>();
		for (var type : MocapType.values()) {
			var averages = KnnPrecision.average(KnnPrecision.compute(
				loadData(dataFileNames(type), type.dataClass())
			));

			typeAverages.put(type, averages);
		}

		var typeDataSeries = typeAverages.entrySet().parallelStream().collect(Collectors.toUnmodifiableMap(
			Entry::getKey,
			entry -> {
				var averages = entry.getValue();
				var averagesDoubles = averages.entrySet().stream()
					.collect(Collectors.toUnmodifiableMap(
						entry1 -> Double.parseDouble(entry1.getKey().id),
						Entry::getValue
					));

				var pointSet = new PointSet(
					averagesDoubles.entrySet().stream().map(Vector2::fromEntry).toList()
				); // for sorting the values
				return pointSet.toDataSeries();
		}));

		var dataset = new DefaultXYDataset();
		typeDataSeries.forEach(dataset::addSeries);

		var chart = ChartFactory.createScatterPlot(
			"average precision for each category for each type",
			"category",
			"average precision",
			dataset,
			PlotOrientation.VERTICAL,
			true,
			false,
			false
		);

		Utils.writeXYGraphAsPdf(
			chart,
			new File(outputPath, "knn averages.pdf"),
			new Vector2<>(1024, 512),
			SmoothXyGraphTheme.instance()
		);
	}

	/**
	 * kNN tests averaged for each category, sorted and to a data series for each mocap type, and to a graph
	 *
	 * @param outputPath where the result shall be written
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestsAveragedOrderedToGraphs(String outputPath) throws IOException {
		// TODO: pro serialisaci použít funkce z Utils
		Map<MocapType, double[][]> typeDataSeries = null;
		var serialisedData = new File("cache/knnTestsAveragedOrderedToGraphs");

		if (serialisedData.canRead()) {
			var deserialiser = new ObjectInputStream(new FileInputStream(serialisedData));
			try {
				typeDataSeries = (Map<MocapType, double[][]>)deserialiser.readObject();
				deserialiser.close();
				System.err.printf("Cached data from %s were read successfully.%n", serialisedData);
			} catch (Exception e) {
				System.err.printf("File with potential cache %s was found, but the data weren't read because of an exception:%n%s%n",
					serialisedData, e);
		}}

		if (typeDataSeries == null) {
			var typeAverages = new HashMap<MocapType, Map<Category, Double>>();
			for (var type : MocapType.values()) {
				var averages = KnnPrecision.average(KnnPrecision.compute(
					loadData(dataFileNames(type), type.dataClass())
				));

				typeAverages.put(type, averages);
			}

			typeDataSeries = typeAverages.entrySet().parallelStream().collect(Collectors.toUnmodifiableMap(
				Entry::getKey,
				entry -> {
					var averages = entry.getValue();
					var averagesDoubles = averages.entrySet().stream()
						.collect(Collectors.toUnmodifiableMap(
							entry1 -> Double.parseDouble(entry1.getKey().id),
							// convert categories to their ids but doubles
							Entry::getValue
						));

					var sortedPoints = new TreeSet<Vector2<Double>>(Comparator.comparing(Vector2::b));
					sortedPoints.addAll(averagesDoubles.entrySet().stream().map(Vector2::fromEntry).toList());
					var bValues = sortedPoints.stream().mapToDouble(Vector2::b).toArray();
					var range = IntStream.range(0, bValues.length)
						.mapToDouble(i -> (double)i).toArray();
					return new double[][]{range, bValues};
				}));

			var serialiser = new ObjectOutputStream(new FileOutputStream(serialisedData));
			try {
				serialiser.writeObject(typeDataSeries);
				serialiser.close();
				System.err.printf("Data cache was written to %s.%n", serialisedData);
			} catch (Exception e) {
				System.err.printf("There was an attempt to write data as cache to %s, but the data writing failed because of an exception:%n%s%n",
					serialisedData, e);
			}
		}

		var dataset = new DefaultXYDataset();
		typeDataSeries.forEach(dataset::addSeries);

		var chart = ChartFactory.createScatterPlot(
			"average precision for each category for each type, with points sorted by their height",
			"",
			"average precision",
			dataset,
			PlotOrientation.VERTICAL,
			true,
			false,
			false
		);

		Utils.writeXYGraphAsPdf(
			chart,
			new File(outputPath, "knn averages.pdf"),
			new Vector2<>(1024, 512),
			SmoothXyGraphTheme.instance()
		);
	}

	private interface ShowGraph {
		void show(String dataPath, Class<? extends LocalAbstractObject> dataClass, String outputPath) throws IOException;
	}

	private static String dataFileNames(MocapType type) {
		return switch (type) {
			case SKELETON -> "data/skeleta/class130-actions-coords_normPOS-fps12.data";
			case HARD_MW -> "data/motion words/hdm05-annotations_specific-segment80_shift16-coords_normPOS-fps12-quantized-pivots-kmedoids-350.data";
			case SOFT_MW -> "data/motion words/kmedoids-350-softassign-D20K6.data";
			case MATCH_N_MW -> "data/motion words/multi-overlay_motion_words.data";
		};
	}
}
